import os

import logbook
import pytest

from .utils import require_module


@require_module('win32con')
@require_module('win32evtlog')
@require_module('win32evtlogutil')
@pytest.mark.skipif(os.environ.get('ENABLE_LOGBOOK_NTEVENTLOG_TESTS') is None,
                    reason="Don't clutter NT Event Log unless enabled.")
def test_nteventlog_handler():
    from win32con import (
        EVENTLOG_ERROR_TYPE, EVENTLOG_INFORMATION_TYPE, EVENTLOG_WARNING_TYPE)
    from win32evtlog import (
        EVENTLOG_BACKWARDS_READ, EVENTLOG_SEQUENTIAL_READ, OpenEventLog,
        ReadEventLog)
    from win32evtlogutil import SafeFormatMessage

    logger = logbook.Logger('Test Logger')

    with logbook.NTEventLogHandler('Logbook Test Suite'):
        logger.info('The info log message.')
        logger.warning('The warning log message.')
        logger.error('The error log message.')

    def iter_event_log(handle, flags, offset):
        while True:
            events = ReadEventLog(handle, flags, offset)
            for event in events:
                yield event
            if not events:
                break

    handle = OpenEventLog(None, 'Application')
    flags = EVENTLOG_BACKWARDS_READ | EVENTLOG_SEQUENTIAL_READ

    for event in iter_event_log(handle, flags, 0):
        source = str(event.SourceName)
        if source == 'Logbook Test Suite':
            message = SafeFormatMessage(event, 'Application')
            if 'Message Level: INFO' in message:
                assert 'The info log message' in message
                assert event.EventType == EVENTLOG_INFORMATION_TYPE
            if 'Message Level: WARNING' in message:
                assert 'The warning log message' in message
                assert event.EventType == EVENTLOG_WARNING_TYPE
            if 'Message Level: ERROR' in message:
                assert 'The error log message' in message
                assert event.EventType == EVENTLOG_ERROR_TYPE



import sys

import logbook
from logbook.helpers import StringIO

import pytest

from .utils import capturing_stderr_context, missing, require_module


@require_module('jinja2')
def test_jinja_formatter(logger):
    from logbook.more import JinjaFormatter
    fmter = JinjaFormatter('{{ record.channel }}/{{ record.level_name }}')
    handler = logbook.TestHandler()
    handler.formatter = fmter
    with handler:
        logger.info('info')
    assert 'testlogger/INFO' in handler.formatted_records


@missing('jinja2')
def test_missing_jinja2():
    from logbook.more import JinjaFormatter
    # check the RuntimeError is raised
    with pytest.raises(RuntimeError):
        JinjaFormatter('dummy')


def test_colorizing_support(logger):
    from logbook.more import ColorizedStderrHandler

    class TestColorizingHandler(ColorizedStderrHandler):
        def __init__(self, *args, **kwargs):
            super(TestColorizingHandler, self).__init__(*args, **kwargs)
            self._obj_stream = StringIO()

        @property
        def stream(self):
            return self._obj_stream

    with TestColorizingHandler(format_string='{record.message}') as handler:
        handler.force_color()
        logger.error('An error')
        logger.warn('A warning')
        logger.debug('A debug message')
        lines = handler.stream.getvalue().rstrip('\n').splitlines()
        assert lines == [
            '\x1b[31;01mAn error\x1b[39;49;00m',
            '\x1b[33;01mA warning\x1b[39;49;00m',
            '\x1b[37mA debug message\x1b[39;49;00m']

    with TestColorizingHandler(format_string='{record.message}') as handler:
        handler.forbid_color()
        logger.error('An error')
        logger.warn('A warning')
        logger.debug('A debug message')
        lines = handler.stream.getvalue().rstrip('\n').splitlines()
        assert lines == ['An error', 'A warning', 'A debug message']



def test_tagged(default_handler):
    from logbook.more import TaggingLogger, TaggingHandler
    stream = StringIO()
    second_handler = logbook.StreamHandler(stream)

    logger = TaggingLogger('name', ['cmd'])
    handler = TaggingHandler(dict(
        info=default_handler,
        cmd=second_handler,
        both=[default_handler, second_handler],
    ))
    handler.bubble = False

    with handler:
        with capturing_stderr_context() as captured:
            logger.log('info', 'info message')
            logger.log('both', 'all message')
            logger.cmd('cmd message')

    stderr = captured.getvalue()

    assert 'info message' in stderr
    assert 'all message' in stderr
    assert 'cmd message' not in stderr

    stringio = stream.getvalue()

    assert 'info message' not in stringio
    assert 'all message' in stringio
    assert 'cmd message' in stringio


def test_tagging_logger(default_handler):
    from logbook import StderrHandler
    from logbook.more import TaggingLogger

    logger = TaggingLogger('tagged', ['a', 'b'])
    handler = StderrHandler(format_string="{record.msg}|{record.extra[tags]}")

    with handler:
        with capturing_stderr_context() as captured:
            logger.a("a")
            logger.b("b")

    stderr = captured.getvalue()

    assert "a|['a']" in stderr
    assert "a|['b']" not in stderr
    assert "b|['b']" in stderr
    assert "b|['a']" not in stderr


def test_external_application_handler(tmpdir, logger):
    from logbook.more import ExternalApplicationHandler as Handler
    fn = tmpdir.join('tempfile')
    handler = Handler([sys.executable, '-c', r'''if 1:
    f = open(%(tempfile)s, 'w')
    try:
        f.write('{record.message}\n')
    finally:
        f.close()
    ''' % {'tempfile': repr(str(fn))}])
    with handler:
        logger.error('this is a really bad idea')
    with fn.open() as rf:
        contents = rf.read().strip()
    assert contents == 'this is a really bad idea'


def test_exception_handler(logger):
    from logbook.more import ExceptionHandler

    with ExceptionHandler(ValueError):
        with pytest.raises(ValueError) as caught:
            logger.info('here i am')
    assert 'INFO: testlogger: here i am' in caught.value.args[0]


def test_exception_handler_specific_level(logger):
    from logbook.more import ExceptionHandler
    with logbook.TestHandler() as test_handler:
        with pytest.raises(ValueError) as caught:
            with ExceptionHandler(ValueError, level='WARNING'):
                logger.info('this is irrelevant')
                logger.warn('here i am')
        assert 'WARNING: testlogger: here i am' in caught.value.args[0]
    assert 'this is irrelevant' in test_handler.records[0].message


def test_dedup_handler(logger):
    from logbook.more import DedupHandler
    with logbook.TestHandler() as test_handler:
        with DedupHandler():
            logger.info('foo')
            logger.info('bar')
            logger.info('foo')
    assert 2 == len(test_handler.records)
    assert 'message repeated 2 times: foo' in test_handler.records[0].message
    assert 'message repeated 1 times: bar' in test_handler.records[1].message


class TestRiemannHandler(object):

    @require_module("riemann_client")
    def test_happy_path(self, logger):
        from logbook.more import RiemannHandler
        riemann_handler = RiemannHandler("127.0.0.1", 5555, message_type="test", level=logbook.INFO)
        null_handler = logbook.NullHandler()
        with null_handler.applicationbound():
            with riemann_handler:
                logger.error("Something bad has happened")
                try:
                    raise RuntimeError("For example, a RuntimeError")
                except Exception as ex:
                    logger.exception(ex)
                logger.info("But now it is ok")

        q = riemann_handler.queue
        assert len(q) == 3
        error_event = q[0]
        assert error_event["state"] == "error"
        exc_event = q[1]
        assert exc_event["description"] == "For example, a RuntimeError"
        info_event = q[2]
        assert info_event["state"] == "ok"

    @require_module("riemann_client")
    def test_incorrect_type(self):
        from logbook.more import RiemannHandler
        with pytest.raises(RuntimeError):
            RiemannHandler("127.0.0.1", 5555, message_type="fancy_type")

    @require_module("riemann_client")
    def test_flush(self, logger):
        from logbook.more import RiemannHandler
        riemann_handler = RiemannHandler("127.0.0.1",
                                         5555,
                                         message_type="test",
                                         flush_threshold=2,
                                         level=logbook.INFO)
        null_handler = logbook.NullHandler()
        with null_handler.applicationbound():
            with riemann_handler:
                logger.info("Msg #1")
                logger.info("Msg #2")
                logger.info("Msg #3")

        q = riemann_handler.queue
        assert len(q) == 1
        assert q[0]["description"] == "Msg #3"



import logbook


def test_groups(logger):
    def inject_extra(record):
        record.extra['foo'] = 'bar'
    group = logbook.LoggerGroup(processor=inject_extra)
    group.level = logbook.ERROR
    group.add_logger(logger)
    with logbook.TestHandler() as handler:
        logger.warn('A warning')
        logger.error('An error')
    assert (not handler.has_warning('A warning'))
    assert handler.has_error('An error')
    assert handler.records[0].extra['foo'] == 'bar'


def test_group_disabled():
    group = logbook.LoggerGroup()
    logger1 = logbook.Logger('testlogger1')
    logger2 = logbook.Logger('testlogger2')

    group.add_logger(logger1)
    group.add_logger(logger2)

    # Test group disable

    group.disable()

    with logbook.TestHandler() as handler:
        logger1.warn('Warning 1')
        logger2.warn('Warning 2')

    assert not handler.has_warnings

    # Test group enable

    group.enable()

    with logbook.TestHandler() as handler:
        logger1.warn('Warning 1')
        logger2.warn('Warning 2')

    assert handler.has_warning('Warning 1')
    assert handler.has_warning('Warning 2')

    # Test group disabled, but logger explicitly enabled

    group.disable()

    logger1.enable()

    with logbook.TestHandler() as handler:
        logger1.warn('Warning 1')
        logger2.warn('Warning 2')

    assert handler.has_warning('Warning 1')
    assert not handler.has_warning('Warning 2')

    # Logger 1 will be enabled by using force=True

    group.disable(force=True)

    with logbook.TestHandler() as handler:
        logger1.warn('Warning 1')
        logger2.warn('Warning 2')

    assert not handler.has_warning('Warning 1')
    assert not handler.has_warning('Warning 2')

    # Enabling without force means logger 1 will still be disabled.

    group.enable()

    with logbook.TestHandler() as handler:
        logger1.warn('Warning 1')
        logger2.warn('Warning 2')

    assert not handler.has_warning('Warning 1')
    assert handler.has_warning('Warning 2')

    # Force logger 1 enabled.

    group.enable(force=True)

    with logbook.TestHandler() as handler:
        logger1.warn('Warning 1')
        logger2.warn('Warning 2')

    assert handler.has_warning('Warning 1')
    assert handler.has_warning('Warning 2')



# -*- coding: utf-8 -*-
"""
    logbook.base
    ~~~~~~~~~~~~

    Base implementation for logbook.

    :copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
    :license: BSD, see LICENSE for more details.
"""
import os
import sys
import traceback
from collections import defaultdict
from datetime import datetime
from itertools import chain
from weakref import ref as weakref

from logbook.concurrency import (greenlet_get_ident, thread_get_ident,
                                 thread_get_name)

from logbook.helpers import (PY2, cached_property, integer_types, iteritems,
                             parse_iso8601, string_types, to_safe_json, u,
                             xrange)

_has_speedups = False
try:
    if os.environ.get('DISABLE_LOGBOOK_CEXT_AT_RUNTIME'):
        raise ImportError("Speedups disabled via DISABLE_LOGBOOK_CEXT_AT_RUNTIME")

    from logbook._speedups import (
        _missing, group_reflected_property, ContextStackManager, StackedObject)

    _has_speedups = True
except ImportError:
    from logbook._fallback import (
        _missing, group_reflected_property, ContextStackManager, StackedObject)

_datetime_factory = datetime.utcnow


def set_datetime_format(datetime_format):
    """
    Set the format for the datetime objects created, which are then
    made available as the :py:attr:`LogRecord.time` attribute of
    :py:class:`LogRecord` instances.

    :param datetime_format: Indicates how to generate datetime objects.

    Possible values are:

         "utc"
             :py:attr:`LogRecord.time` will be a datetime in UTC time zone
             (but not time zone aware)
         "local"
             :py:attr:`LogRecord.time` will be a datetime in local time zone
             (but not time zone aware)
         A `callable` returning datetime instances
            :py:attr:`LogRecord.time` will be a datetime created by
            :py:obj:`datetime_format` (possibly time zone aware)

    This function defaults to creating datetime objects in UTC time,
    using `datetime.utcnow()
    <http://docs.python.org/3/library/datetime.html#datetime.datetime.utcnow>`_,
    so that logbook logs all times in UTC time by default.  This is
    recommended in case you have multiple software modules or
    instances running in different servers in different time zones, as
    it makes it simple and less error prone to correlate logging
    across the different servers.

    On the other hand if all your software modules are running in the
    same time zone and you have to correlate logging with third party
    modules already logging in local time, it can be more convenient
    to have logbook logging to local time instead of UTC.  Local time
    logging can be enabled like this::

       import logbook
       from datetime import datetime
       logbook.set_datetime_format("local")

    Other uses rely on your supplied :py:obj:`datetime_format`.
    Using `pytz <https://pypi.org/pypi/pytz>`_ for example::

        from datetime import datetime
        import logbook
        import pytz

        def utc_tz():
            return datetime.now(tz=pytz.utc)

        logbook.set_datetime_format(utc_tz)
    """
    global _datetime_factory
    if datetime_format == "utc":
        _datetime_factory = datetime.utcnow
    elif datetime_format == "local":
        _datetime_factory = datetime.now
    elif callable(datetime_format):
        inst = datetime_format()
        if not isinstance(inst, datetime):
            raise ValueError("Invalid callable value, valid callable "
                             "should return datetime.datetime instances, "
                             "not %r" % (type(inst),))
        _datetime_factory = datetime_format
    else:
        raise ValueError("Invalid value %r.  Valid values are 'utc' and "
                         "'local'." % (datetime_format,))

# make sure to sync these up with _speedups.pyx
CRITICAL = 15
ERROR = 14
WARNING = 13
NOTICE = 12
INFO = 11
DEBUG = 10
TRACE = 9
NOTSET = 0

_level_names = {
    CRITICAL:   'CRITICAL',
    ERROR:      'ERROR',
    WARNING:    'WARNING',
    NOTICE:     'NOTICE',
    INFO:       'INFO',
    DEBUG:      'DEBUG',
    TRACE:      'TRACE',
    NOTSET:     'NOTSET'
}
_reverse_level_names = dict((v, k) for (k, v) in iteritems(_level_names))
_missing = object()


# on python 3 we can savely assume that frame filenames will be in
# unicode, on Python 2 we have to apply a trick.
if PY2:
    def _convert_frame_filename(fn):
        if isinstance(fn, unicode):
            fn = fn.decode(sys.getfilesystemencoding() or 'utf-8',
                           'replace')
        return fn
else:
    def _convert_frame_filename(fn):
        return fn


def level_name_property():
    """Returns a property that reflects the level as name from
    the internal level attribute.
    """

    def _get_level_name(self):
        return get_level_name(self.level)

    def _set_level_name(self, level):
        self.level = lookup_level(level)
    return property(_get_level_name, _set_level_name,
                    doc='The level as unicode string')


def lookup_level(level):
    """Return the integer representation of a logging level."""
    if isinstance(level, integer_types):
        return level
    try:
        return _reverse_level_names[level]
    except KeyError:
        raise LookupError('unknown level name %s' % level)


def get_level_name(level):
    """Return the textual representation of logging level 'level'."""
    try:
        return _level_names[level]
    except KeyError:
        raise LookupError('unknown level')


class _ExceptionCatcher(object):
    """Helper for exception caught blocks."""

    def __init__(self, logger, args, kwargs):
        self.logger = logger
        self.args = args
        self.kwargs = kwargs

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            kwargs = self.kwargs.copy()
            kwargs['exc_info'] = (exc_type, exc_value, tb)
            self.logger.exception(*self.args, **kwargs)
        return True


class ContextObject(StackedObject):
    """An object that can be bound to a context.  It is managed by the
    :class:`ContextStackManager`"""

    #: subclasses have to instanciate a :class:`ContextStackManager`
    #: object on this attribute which is then shared for all the
    #: subclasses of it.
    stack_manager = None

    def push_greenlet(self):
        """Pushes the context object to the greenlet stack."""
        self.stack_manager.push_greenlet(self)

    def pop_greenlet(self):
        """Pops the context object from the stack."""
        popped = self.stack_manager.pop_greenlet()
        assert popped is self, 'popped unexpected object'

    def push_context(self):
        """Pushes the context object to the context stack."""
        self.stack_manager.push_context(self)

    def pop_context(self):
        """Pops the context object from the stack."""
        popped = self.stack_manager.pop_context()
        assert popped is self, 'popped unexpected object'

    def push_thread(self):
        """Pushes the context object to the thread stack."""
        self.stack_manager.push_thread(self)

    def pop_thread(self):
        """Pops the context object from the stack."""
        popped = self.stack_manager.pop_thread()
        assert popped is self, 'popped unexpected object'

    def push_application(self):
        """Pushes the context object to the application stack."""
        self.stack_manager.push_application(self)

    def pop_application(self):
        """Pops the context object from the stack."""
        popped = self.stack_manager.pop_application()
        assert popped is self, 'popped unexpected object'


class NestedSetup(StackedObject):
    """A nested setup can be used to configure multiple handlers
    and processors at once.
    """

    def __init__(self, objects=None):
        self.objects = list(objects or ())

    def push_application(self):
        for obj in self.objects:
            obj.push_application()

    def pop_application(self):
        for obj in reversed(self.objects):
            obj.pop_application()

    def push_thread(self):
        for obj in self.objects:
            obj.push_thread()

    def pop_thread(self):
        for obj in reversed(self.objects):
            obj.pop_thread()

    def push_greenlet(self):
        for obj in self.objects:
            obj.push_greenlet()

    def pop_greenlet(self):
        for obj in reversed(self.objects):
            obj.pop_greenlet()

    def push_context(self):
        for obj in self.objects:
            obj.push_context()

    def pop_context(self):
        for obj in reversed(self.objects):
            obj.pop_context()


class Processor(ContextObject):
    """Can be pushed to a stack to inject additional information into
    a log record as necessary::

        def inject_ip(record):
            record.extra['ip'] = '127.0.0.1'

        with Processor(inject_ip):
            ...
    """

    stack_manager = ContextStackManager()

    def __init__(self, callback=None):
        #: the callback that was passed to the constructor
        self.callback = callback

    def process(self, record):
        """Called with the log record that should be overridden.  The default
        implementation calls :attr:`callback` if it is not `None`.
        """
        if self.callback is not None:
            self.callback(record)


class _InheritedType(object):
    __slots__ = ()

    def __repr__(self):
        return 'Inherit'

    def __reduce__(self):
        return 'Inherit'
Inherit = _InheritedType()


class Flags(ContextObject):
    """Allows flags to be pushed on a flag stack.  Currently two flags
    are available:

    `errors`
        Can be set to override the current error behaviour.  This value is
        used when logging calls fail.  The default behaviour is spitting
        out the stacktrace to stderr but this can be overridden:

        =================== ==========================================
        ``'silent'``        fail silently
        ``'raise'``         raise a catchable exception
        ``'print'``         print the stacktrace to stderr (default)
        =================== ==========================================

    `introspection`
        Can be used to disable frame introspection.  This can give a
        speedup on production systems if you are using a JIT compiled
        Python interpreter such as pypy.  The default is `True`.

        Note that the default setup of some of the handler (mail for
        instance) includes frame dependent information which will
        not be available when introspection is disabled.

    Example usage::

        with Flags(errors='silent'):
            ...
    """
    stack_manager = ContextStackManager()

    def __init__(self, **flags):
        self.__dict__.update(flags)

    @staticmethod
    def get_flag(flag, default=None):
        """Looks up the current value of a specific flag."""
        for flags in Flags.stack_manager.iter_context_objects():
            val = getattr(flags, flag, Inherit)
            if val is not Inherit:
                return val
        return default


def _create_log_record(cls, dict):
    """Extra function for reduce because on Python 3 unbound methods
    can no longer be pickled.
    """
    return cls.from_dict(dict)


class LogRecord(object):
    """A LogRecord instance represents an event being logged.

    LogRecord instances are created every time something is logged. They
    contain all the information pertinent to the event being logged. The
    main information passed in is in msg and args
    """
    _pullable_information = frozenset((
        'func_name', 'module', 'filename', 'lineno', 'process_name', 'thread',
        'thread_name', 'greenlet', 'formatted_exception', 'message',
        'exception_name', 'exception_message'
    ))
    _noned_on_close = frozenset(('exc_info', 'frame', 'calling_frame'))

    #: can be overriden by a handler to not close the record.  This could
    #: lead to memory leaks so it should be used carefully.
    keep_open = False

    #: the time of the log record creation as :class:`datetime.datetime`
    #: object.  This information is unavailable until the record was
    #: heavy initialized.
    time = None

    #: a flag that is `True` if the log record is heavy initialized which
    #: is not the case by default.
    heavy_initialized = False

    #: a flag that is `True` when heavy initialization is no longer possible
    late = False

    #: a flag that is `True` when all the information was pulled from the
    #: information that becomes unavailable on close.
    information_pulled = False

    def __init__(self, channel, level, msg, args=None, kwargs=None,
                 exc_info=None, extra=None, frame=None, dispatcher=None,
                 frame_correction=0):
        #: the name of the logger that created it or any other textual
        #: channel description.  This is a descriptive name and can be
        #: used for filtering.
        self.channel = channel
        #: The message of the log record as new-style format string.
        self.msg = msg
        #: the positional arguments for the format string.
        self.args = args or ()
        #: the keyword arguments for the format string.
        self.kwargs = kwargs or {}
        #: the level of the log record as integer.
        self.level = level
        #: optional exception information.  If set, this is a tuple in the
        #: form ``(exc_type, exc_value, tb)`` as returned by
        #: :func:`sys.exc_info`.
        #: This parameter can also be ``True``, which would cause the exception
        #: info tuple to be fetched for you.
        if not exc_info:
            # this is a special case where exc_info=False can be passed in
            # theory, and it should be the same as exc_info=None
            exc_info = None
        self.exc_info = exc_info
        #: optional extra information as dictionary.  This is the place
        #: where custom log processors can attach custom context sensitive
        #: data.

        # TODO: Replace the lambda with str when we remove support for python 2
        self.extra = defaultdict(lambda: u'', extra or ())
        #: If available, optionally the interpreter frame that pulled the
        #: heavy init.  This usually points to somewhere in the dispatcher.
        #: Might not be available for all calls and is removed when the log
        #: record is closed.
        self.frame = frame
        #: A positive integer telling the number of frames to go back from
        #: the frame which triggered the log entry. This is mainly useful
        #: for decorators that want to show that the log was emitted from
        #: form the function they decorate
        self.frame_correction = frame_correction
        #: the PID of the current process
        self.process = None
        if dispatcher is not None:
            dispatcher = weakref(dispatcher)
        self._dispatcher = dispatcher

    def heavy_init(self):
        """Does the heavy initialization that could be expensive.  This must
        not be called from a higher stack level than when the log record was
        created and the later the initialization happens, the more off the
        date information will be for example.

        This is internally used by the record dispatching system and usually
        something not to worry about.
        """
        if self.heavy_initialized:
            return
        assert not self.late, 'heavy init is no longer possible'
        self.heavy_initialized = True
        self.process = os.getpid()
        self.time = _datetime_factory()
        if self.frame is None and Flags.get_flag('introspection', True):
            self.frame = sys._getframe(1)
        if self.exc_info is True:
            self.exc_info = sys.exc_info()

    def pull_information(self):
        """A helper function that pulls all frame-related information into
        the object so that this information is available after the log
        record was closed.
        """
        if self.information_pulled:
            return
        # due to how cached_property is implemented, the attribute access
        # has the side effect of caching the attribute on the instance of
        # the class.
        for key in self._pullable_information:
            getattr(self, key)
        self.information_pulled = True

    def close(self):
        """Closes the log record.  This will set the frame and calling
        frame to `None` and frame-related information will no longer be
        available unless it was pulled in first (:meth:`pull_information`).
        This makes a log record safe for pickling and will clean up
        memory that might be still referenced by the frames.
        """
        for key in self._noned_on_close:
            setattr(self, key, None)
        self.late = True

    def __reduce_ex__(self, protocol):
        return _create_log_record, (type(self), self.to_dict())

    def to_dict(self, json_safe=False):
        """Exports the log record into a dictionary without the information
        that cannot be safely serialized like interpreter frames and
        tracebacks.
        """
        self.pull_information()
        rv = {}
        for key, value in iteritems(self.__dict__):
            if key[:1] != '_' and key not in self._noned_on_close:
                rv[key] = value
        # the extra dict is exported as regular dict
        rv['extra'] = dict(rv['extra'])
        if json_safe:
            return to_safe_json(rv)
        return rv

    @classmethod
    def from_dict(cls, d):
        """Creates a log record from an exported dictionary.  This also
        supports JSON exported dictionaries.
        """
        rv = object.__new__(cls)
        rv.update_from_dict(d)
        return rv

    def update_from_dict(self, d):
        """Like the :meth:`from_dict` classmethod, but will update the
        instance in place.  Helpful for constructors.
        """
        self.__dict__.update(d)
        for key in self._noned_on_close:
            setattr(self, key, None)
        self._information_pulled = True
        self._channel = None
        if isinstance(self.time, string_types):
            self.time = parse_iso8601(self.time)

        # TODO: Replace the lambda with str when we remove support for python 2`
        self.extra = defaultdict(lambda: u'', self.extra)
        return self

    def _format_message(self, msg, *args, **kwargs):
        """Called if the record's message needs to be formatted.
        Subclasses can implement their own formatting.
        """
        return msg.format(*args, **kwargs)

    @cached_property
    def message(self):
        """The formatted message."""
        if not (self.args or self.kwargs):
            return self.msg
        try:
            try:
                return self._format_message(self.msg, *self.args,
                                            **self.kwargs)
            except UnicodeDecodeError:
                # Assume an unicode message but mixed-up args
                msg = self.msg.encode('utf-8', 'replace')
                return self._format_message(msg, *self.args, **self.kwargs)
            except (UnicodeEncodeError, AttributeError):
                # we catch AttributeError since if msg is bytes,
                # it won't have the 'format' method
                if (sys.exc_info()[0] is AttributeError
                        and (PY2 or not isinstance(self.msg, bytes))):
                    # this is not the case we thought it is...
                    raise
                # Assume encoded message with unicode args.
                # The assumption of utf8 as input encoding is just a guess,
                # but this codepath is unlikely (if the message is a constant
                # string in the caller's source file)
                msg = self.msg.decode('utf-8', 'replace')
                return self._format_message(msg, *self.args, **self.kwargs)

        except Exception:
            # this obviously will not give a proper error message if the
            # information was not pulled and the log record no longer has
            # access to the frame.  But there is not much we can do about
            # that.
            e = sys.exc_info()[1]
            errormsg = ('Could not format message with provided '
                        'arguments: {err}\n  msg={msg!r}\n  '
                        'args={args!r} \n  kwargs={kwargs!r}.\n'
                        'Happened in file {file}, line {lineno}').format(
                err=e, msg=self.msg, args=self.args,
                kwargs=self.kwargs, file=self.filename,
                lineno=self.lineno
            )
            if PY2:
                errormsg = errormsg.encode('utf-8')
            raise TypeError(errormsg)

    level_name = level_name_property()

    @cached_property
    def calling_frame(self):
        """The frame in which the record has been created.  This only
        exists for as long the log record is not closed.
        """
        frm = self.frame
        globs = globals()
        while frm is not None and frm.f_globals is globs:
            frm = frm.f_back

        for _ in xrange(self.frame_correction):
            if frm is None:
                break

            frm = frm.f_back

        return frm

    @cached_property
    def func_name(self):
        """The name of the function that triggered the log call if
        available.  Requires a frame or that :meth:`pull_information`
        was called before.
        """
        cf = self.calling_frame
        if cf is not None:
            return cf.f_code.co_name

    @cached_property
    def module(self):
        """The name of the module that triggered the log call if
        available.  Requires a frame or that :meth:`pull_information`
        was called before.
        """
        cf = self.calling_frame
        if cf is not None:
            return cf.f_globals.get('__name__')

    @cached_property
    def filename(self):
        """The filename of the module in which the record has been created.
        Requires a frame or that :meth:`pull_information` was called before.
        """
        cf = self.calling_frame
        if cf is not None:
            fn = cf.f_code.co_filename
            if fn[:1] == '<' and fn[-1:] == '>':
                return fn
            return _convert_frame_filename(os.path.abspath(fn))

    @cached_property
    def lineno(self):
        """The line number of the file in which the record has been created.
        Requires a frame or that :meth:`pull_information` was called before.
        """
        cf = self.calling_frame
        if cf is not None:
            return cf.f_lineno

    @cached_property
    def greenlet(self):
        """The ident of the greenlet.  This is evaluated late and means that
        if the log record is passed to another greenlet,
        :meth:`pull_information` was called in the old greenlet.
        """
        return greenlet_get_ident()

    @cached_property
    def thread(self):
        """The ident of the thread.  This is evaluated late and means that
        if the log record is passed to another thread, :meth:`pull_information`
        was called in the old thread.
        """
        return thread_get_ident()

    @cached_property
    def thread_name(self):
        """The name of the thread.  This is evaluated late and means that
        if the log record is passed to another thread, :meth:`pull_information`
        was called in the old thread.
        """
        return thread_get_name()

    @cached_property
    def process_name(self):
        """The name of the process in which the record has been created."""
        # Errors may occur if multiprocessing has not finished loading
        # yet - e.g. if a custom import hook causes third-party code
        # to run when multiprocessing calls import. See issue 8200
        # for an example
        mp = sys.modules.get('multiprocessing')
        if mp is not None:  # pragma: no cover
            try:
                return mp.current_process().name
            except Exception:
                pass

    @cached_property
    def formatted_exception(self):
        """The formatted exception which caused this record to be created
        in case there was any.
        """
        if self.exc_info is not None and self.exc_info != (None, None, None):
            rv = ''.join(traceback.format_exception(*self.exc_info))
            if PY2:
                rv = rv.decode('utf-8', 'replace')
            return rv.rstrip()

    @cached_property
    def exception_name(self):
        """The name of the exception."""
        if self.exc_info is not None:
            cls = self.exc_info[0]
            return u(cls.__module__ + '.' + cls.__name__)

    @property
    def exception_shortname(self):
        """An abbreviated exception name (no import path)"""
        return self.exception_name.rsplit('.')[-1]

    @cached_property
    def exception_message(self):
        """The message of the exception."""
        if self.exc_info is not None:
            val = self.exc_info[1]
            try:
                if PY2:
                    return unicode(val)
                else:
                    return str(val)
            except UnicodeError:
                return str(val).decode('utf-8', 'replace')

    @property
    def dispatcher(self):
        """The dispatcher that created the log record.  Might not exist because
        a log record does not have to be created from a logger or other
        dispatcher to be handled by logbook.  If this is set, it will point to
        an object that implements the :class:`~logbook.base.RecordDispatcher`
        interface.
        """
        if self._dispatcher is not None:
            return self._dispatcher()


class LoggerMixin(object):
    """This mixin class defines and implements the "usual" logger
    interface (i.e. the descriptive logging functions).

    Classes using this mixin have to implement a :meth:`!handle` method which
    takes a :class:`~logbook.LogRecord` and passes it along.
    """

    #: The name of the minimium logging level required for records to be
    #: created.
    level_name = level_name_property()

    def trace(self, *args, **kwargs):
        """Logs a :class:`~logbook.LogRecord` with the level set
        to :data:`~logbook.TRACE`.
        """
        if not self.disabled and TRACE >= self.level:
            self._log(TRACE, args, kwargs)

    def debug(self, *args, **kwargs):
        """Logs a :class:`~logbook.LogRecord` with the level set
        to :data:`~logbook.DEBUG`.
        """
        if not self.disabled and DEBUG >= self.level:
            self._log(DEBUG, args, kwargs)

    def info(self, *args, **kwargs):
        """Logs a :class:`~logbook.LogRecord` with the level set
        to :data:`~logbook.INFO`.
        """
        if not self.disabled and INFO >= self.level:
            self._log(INFO, args, kwargs)

    def warn(self, *args, **kwargs):
        """Logs a :class:`~logbook.LogRecord` with the level set
        to :data:`~logbook.WARNING`.  This function has an alias
        named :meth:`warning`.
        """
        if not self.disabled and WARNING >= self.level:
            self._log(WARNING, args, kwargs)

    def warning(self, *args, **kwargs):
        """Alias for :meth:`warn`."""
        return self.warn(*args, **kwargs)

    def notice(self, *args, **kwargs):
        """Logs a :class:`~logbook.LogRecord` with the level set
        to :data:`~logbook.NOTICE`.
        """
        if not self.disabled and NOTICE >= self.level:
            self._log(NOTICE, args, kwargs)

    def error(self, *args, **kwargs):
        """Logs a :class:`~logbook.LogRecord` with the level set
        to :data:`~logbook.ERROR`.
        """
        if not self.disabled and ERROR >= self.level:
            self._log(ERROR, args, kwargs)

    def exception(self, *args, **kwargs):
        """Works exactly like :meth:`error` just that the message
        is optional and exception information is recorded.
        """
        if self.disabled or ERROR < self.level:
            return
        if not args:
            args = ('Uncaught exception occurred',)
        if 'exc_info' not in kwargs:
            exc_info = sys.exc_info()
            assert exc_info[0] is not None, 'no exception occurred'
            kwargs.setdefault('exc_info', sys.exc_info())
        return self.error(*args, **kwargs)

    def critical(self, *args, **kwargs):
        """Logs a :class:`~logbook.LogRecord` with the level set
        to :data:`~logbook.CRITICAL`.
        """
        if not self.disabled and CRITICAL >= self.level:
            self._log(CRITICAL, args, kwargs)

    def log(self, level, *args, **kwargs):
        """Logs a :class:`~logbook.LogRecord` with the level set
        to the `level` parameter.  Because custom levels are not
        supported by logbook, this method is mainly used to avoid
        the use of reflection (e.g.: :func:`getattr`) for programmatic
        logging.
        """
        level = lookup_level(level)
        if level >= self.level:
            self._log(level, args, kwargs)

    def catch_exceptions(self, *args, **kwargs):
        """A context manager that catches exceptions and calls
        :meth:`exception` for exceptions caught that way.  Example:

        .. code-block:: python

            with logger.catch_exceptions():
                execute_code_that_might_fail()
        """
        if not args:
            args = ('Uncaught exception occurred',)
        return _ExceptionCatcher(self, args, kwargs)

    def enable(self):
        """Convenience method to enable this logger.

        :raises AttributeError: The disabled property is read-only, typically
                                because it was overridden in a subclass.

        .. versionadded:: 1.0
        """
        try:
            self.disabled = False
        except AttributeError:
            raise AttributeError('The disabled property is read-only.')

    def disable(self):
        """Convenience method to disable this logger.

        :raises AttributeError: The disabled property is read-only, typically
                                because it was overridden in a subclass.

        .. versionadded:: 1.0
        """
        try:
            self.disabled = True
        except AttributeError:
            raise AttributeError('The disabled property is read-only.')

    def _log(self, level, args, kwargs):
        exc_info = kwargs.pop('exc_info', None)
        extra = kwargs.pop('extra', None)
        frame_correction = kwargs.pop('frame_correction', 0)
        self.make_record_and_handle(level, args[0], args[1:], kwargs,
                                    exc_info, extra, frame_correction)


class RecordDispatcher(object):
    """A record dispatcher is the internal base class that implements
    the logic used by the :class:`~logbook.Logger`.
    """

    #: If this is set to `True` the dispatcher information will be suppressed
    #: for log records emitted from this logger.
    suppress_dispatcher = False

    def __init__(self, name=None, level=NOTSET):
        #: the name of the record dispatcher
        self.name = name
        #: list of handlers specific for this record dispatcher
        self.handlers = []
        #: optionally the name of the group this logger belongs to
        self.group = None
        #: the level of the record dispatcher as integer
        self.level = level

    disabled = group_reflected_property('disabled', False)
    level = group_reflected_property('level', NOTSET, fallback=NOTSET)

    def handle(self, record):
        """Call the handlers for the specified record.  This is
        invoked automatically when a record should be handled.
        The default implementation checks if the dispatcher is disabled
        and if the record level is greater than the level of the
        record dispatcher.  In that case it will call the handlers
        (:meth:`call_handlers`).
        """
        if not self.disabled and record.level >= self.level:
            self.call_handlers(record)

    def make_record_and_handle(self, level, msg, args, kwargs, exc_info,
                               extra, frame_correction):
        """Creates a record from some given arguments and heads it
        over to the handling system.
        """
        # The channel information can be useful for some use cases which is
        # why we keep it on there.  The log record however internally will
        # only store a weak reference to the channel, so it might disappear
        # from one instruction to the other.  It will also disappear when
        # a log record is transmitted to another process etc.
        channel = None
        if not self.suppress_dispatcher:
            channel = self

        record = LogRecord(self.name, level, msg, args, kwargs, exc_info,
                           extra, None, channel, frame_correction)

        # after handling the log record is closed which will remove some
        # referenes that would require a GC run on cpython.  This includes
        # the current stack frame, exception information.  However there are
        # some use cases in keeping the records open for a little longer.
        # For example the test handler keeps log records open until the
        # test handler is closed to allow assertions based on stack frames
        # and exception information.
        try:
            self.handle(record)
        finally:
            record.late = True
            if not record.keep_open:
                record.close()

    def call_handlers(self, record):
        """Pass a record to all relevant handlers in the following
        order:

        -   per-dispatcher handlers are handled first
        -   afterwards all the current context handlers in the
            order they were pushed

        Before the first handler is invoked, the record is processed
        (:meth:`process_record`).
        """
        # for performance reasons records are only heavy initialized
        # and processed if at least one of the handlers has a higher
        # level than the record and that handler is not a black hole.
        record_initialized = False

        # Both logger attached handlers as well as context specific
        # handlers are handled one after another.  The latter also
        # include global handlers.
        for handler in chain(self.handlers,
                             Handler.stack_manager.iter_context_objects()):
            # skip records that this handler is not interested in based
            # on the record and handler level or in case this method was
            # overridden on some custom logic.
            if not handler.should_handle(record):
                continue

            # first case of blackhole (without filter).
            # this should discard all further processing and
            # we don't have to heavy_init to know that...
            if handler.filter is None and handler.blackhole:
                break

            # we are about to handle the record.  If it was not yet
            # processed by context-specific record processors we
            # have to do that now and remeber that we processed
            # the record already.
            if not record_initialized:
                record.heavy_init()
                self.process_record(record)
                record_initialized = True

            # a filter can still veto the handling of the record.  This
            # however is already operating on an initialized and processed
            # record.  The impact is that filters are slower than the
            # handler's should_handle function in case there is no default
            # handler that would handle the record (delayed init).
            if (handler.filter is not None
                    and not handler.filter(record, handler)):
                continue

            # We might have a filter, so now that we know we *should* handle
            # this record, we should consider the case of us being a black hole...
            if handler.blackhole:
                break

            # handle the record.  If the record was handled and
            # the record is not bubbling we can abort now.
            if handler.handle(record) and not handler.bubble:
                break

    def process_record(self, record):
        """Processes the record with all context specific processors.  This
        can be overriden to also inject additional information as necessary
        that can be provided by this record dispatcher.
        """
        if self.group is not None:
            self.group.process_record(record)
        for processor in Processor.stack_manager.iter_context_objects():
            processor.process(record)


class Logger(RecordDispatcher, LoggerMixin):
    """Instances of the Logger class represent a single logging channel.
    A "logging channel" indicates an area of an application. Exactly
    how an "area" is defined is up to the application developer.

    Names used by logbook should be descriptive and are intended for user
    display, not for filtering.  Filtering should happen based on the
    context information instead.

    A logger internally is a subclass of a
    :class:`~logbook.base.RecordDispatcher` that implements the actual
    logic.  If you want to implement a custom logger class, have a look
    at the interface of that class as well.
    """


class LoggerGroup(object):
    """A LoggerGroup represents a group of loggers.  It cannot emit log
    messages on its own but it can be used to set the disabled flag and
    log level of all loggers in the group.

    Furthermore the :meth:`process_record` method of the group is called
    by any logger in the group which by default calls into the
    :attr:`processor` callback function.
    """

    def __init__(self, loggers=None, level=NOTSET, processor=None):
        #: a list of all loggers on the logger group.  Use the
        #: :meth:`add_logger` and :meth:`remove_logger` methods to add
        #: or remove loggers from this list.
        self.loggers = []
        if loggers is not None:
            for logger in loggers:
                self.add_logger(logger)

        #: the level of the group.  This is reflected to the loggers
        #: in the group unless they overrode the setting.
        self.level = lookup_level(level)
        #: the disabled flag for all loggers in the group, unless
        #: the loggers overrode the setting.
        self.disabled = False
        #: an optional callback function that is executed to process
        #: the log records of all loggers in the group.
        self.processor = processor

    def add_logger(self, logger):
        """Adds a logger to this group."""
        assert logger.group is None, 'Logger already belongs to a group'
        logger.group = self
        self.loggers.append(logger)

    def remove_logger(self, logger):
        """Removes a logger from the group."""
        self.loggers.remove(logger)
        logger.group = None

    def process_record(self, record):
        """Like :meth:`Logger.process_record` but for all loggers in
        the group.  By default this calls into the :attr:`processor`
        function is it's not `None`.
        """
        if self.processor is not None:
            self.processor(record)

    def enable(self, force=False):
        """Convenience method to enable this group.

        :param force: Force enable loggers that were explicitly set.

        :raises AttributeError: If ``force=True`` and the disabled property of
                                a logger is read-only, typically because it was
                                overridden in a subclass.

        .. versionadded:: 1.0
        """
        self.disabled = False
        if force:
            for logger in self.loggers:
                rv = getattr(logger, '_disabled', _missing)
                if rv is not _missing:
                    logger.enable()

    def disable(self, force=False):
        """Convenience method to disable this group.

        :param force: Force disable loggers that were explicitly set.

        :raises AttributeError: If ``force=True`` and the disabled property of
                                a logger is read-only, typically because it was
                                overridden in a subclass.

        .. versionadded:: 1.0
        """
        self.disabled = True
        if force:
            for logger in self.loggers:
                rv = getattr(logger, '_disabled', _missing)
                if rv is not _missing:
                    logger.disable()


_default_dispatcher = RecordDispatcher()


def dispatch_record(record):
    """Passes a record on to the handlers on the stack.  This is useful when
    log records are created programmatically and already have all the
    information attached and should be dispatched independent of a logger.
    """
    _default_dispatcher.call_handlers(record)

# at that point we are safe to import handler
from logbook.handlers import Handler # isort:skip



import logbook

from .utils import capturing_stderr_context


def test_null_handler(activation_strategy, logger):
    with capturing_stderr_context() as captured:
        with activation_strategy(logbook.NullHandler()):
            with activation_strategy(logbook.TestHandler(level='ERROR')) as handler:
                logger.error('An error')
                logger.warn('A warning')
        assert captured.getvalue() == ''
        assert (not handler.has_warning('A warning'))
        assert handler.has_error('An error')


def test_blackhole_setting(activation_strategy):
    null_handler = logbook.NullHandler()
    heavy_init = logbook.LogRecord.heavy_init
    with activation_strategy(null_handler):
        def new_heavy_init(self):
            raise RuntimeError('should not be triggered')
        logbook.LogRecord.heavy_init = new_heavy_init
        try:
            with activation_strategy(null_handler):
                logbook.warn('Awesome')
        finally:
            logbook.LogRecord.heavy_init = heavy_init

    null_handler.bubble = True
    with capturing_stderr_context() as captured:
        logbook.warning('Not a blockhole')
        assert captured.getvalue() != ''


def test_null_handler_filtering(activation_strategy):
    logger1 = logbook.Logger("1")
    logger2 = logbook.Logger("2")
    outer = logbook.TestHandler()
    inner = logbook.NullHandler()

    inner.filter = lambda record, handler: record.dispatcher is logger1

    with activation_strategy(outer):
        with activation_strategy(inner):
            logger1.warn("1")
            logger2.warn("2")

    assert outer.has_warning('2', channel='2')
    assert (not outer.has_warning('1', channel='1'))



import logbook

from .utils import capturing_stderr_context, make_fake_mail_handler


def test_custom_logger(activation_strategy, logger):
    client_ip = '127.0.0.1'

    class CustomLogger(logbook.Logger):

        def process_record(self, record):
            record.extra['ip'] = client_ip

    custom_log = CustomLogger('awesome logger')
    fmt = ('[{record.level_name}] {record.channel}: '
           '{record.message} [{record.extra[ip]}]')
    handler = logbook.TestHandler(format_string=fmt)
    assert handler.format_string == fmt

    with activation_strategy(handler):
        custom_log.warn('Too many sounds')
        logger.warn('"Music" playing')

    assert handler.formatted_records == [
        '[WARNING] awesome logger: Too many sounds [127.0.0.1]',
        '[WARNING] testlogger: "Music" playing []']


def test_custom_handling(activation_strategy, logger):
    class MyTestHandler(logbook.TestHandler):
        def handle(self, record):
            if record.extra.get('flag') != 'testing':
                return False
            return logbook.TestHandler.handle(self, record)

    # Check metaclass (== cls.__class__)
    assert logbook.Handler.__class__ == logbook.handlers._HandlerType

    class MyLogger(logbook.Logger):
        def process_record(self, record):
            logbook.Logger.process_record(self, record)
            record.extra['flag'] = 'testing'

    log = MyLogger()
    handler = MyTestHandler()
    with capturing_stderr_context() as captured:
        with activation_strategy(handler):
            log.warn('From my logger')
            logger.warn('From another logger')
        assert handler.has_warning('From my logger')
        assert 'From another logger' in captured.getvalue()


def test_nested_setups(activation_strategy):
    with capturing_stderr_context() as captured:
        logger = logbook.Logger('App')
        test_handler = logbook.TestHandler(level='WARNING')
        mail_handler = make_fake_mail_handler(bubble=True)

        handlers = logbook.NestedSetup([
            logbook.NullHandler(),
            test_handler,
            mail_handler
        ])

        with activation_strategy(handlers):
            logger.warn('This is a warning')
            logger.error('This is also a mail')
            try:
                1 / 0
            except Exception:
                logger.exception()
        logger.warn('And here we go straight back to stderr')

        assert test_handler.has_warning('This is a warning')
        assert test_handler.has_error('This is also a mail')
        assert len(mail_handler.mails) == 2
        assert 'This is also a mail' in mail_handler.mails[0][2]
        assert '1 / 0' in mail_handler.mails[1][2]
        assert 'And here we go straight back to stderr' in captured.getvalue()

        with activation_strategy(handlers):
            logger.warn('threadbound warning')

        handlers.push_application()
        try:
            logger.warn('applicationbound warning')
        finally:
            handlers.pop_application()


def test_filtering(activation_strategy):
    logger1 = logbook.Logger('Logger1')
    logger2 = logbook.Logger('Logger2')
    handler = logbook.TestHandler()
    outer_handler = logbook.TestHandler()

    def only_1(record, handler):
        return record.dispatcher is logger1
    handler.filter = only_1

    with activation_strategy(outer_handler):
        with activation_strategy(handler):
            logger1.warn('foo')
            logger2.warn('bar')

    assert handler.has_warning('foo', channel='Logger1')
    assert (not handler.has_warning('bar', channel='Logger2'))
    assert (not outer_handler.has_warning('foo', channel='Logger1'))
    assert outer_handler.has_warning('bar', channel='Logger2')


def test_different_context_pushing(activation_strategy):
    h1 = logbook.TestHandler(level=logbook.DEBUG)
    h2 = logbook.TestHandler(level=logbook.INFO)
    h3 = logbook.TestHandler(level=logbook.WARNING)
    logger = logbook.Logger('Testing')

    with activation_strategy(h1):
        with activation_strategy(h2):
            with activation_strategy(h3):
                logger.warn('Wuuu')
                logger.info('still awesome')
                logger.debug('puzzled')

    assert h1.has_debug('puzzled')
    assert h2.has_info('still awesome')
    assert h3.has_warning('Wuuu')
    for handler in h1, h2, h3:
        assert len(handler.records) == 1


def test_default_handlers(logger):
    with capturing_stderr_context() as stream:
        logger.warn('Aha!')
        captured = stream.getvalue()
    assert 'WARNING: testlogger: Aha!' in captured



import re

import pytest


@pytest.mark.parametrize("level, method", [
    ("trace", "has_traces"),
    ("debug", "has_debugs"),
    ("info", "has_infos"),
    ("notice", "has_notices"),
    ("warning", "has_warnings"),
    ("error", "has_errors"),
    ("critical", "has_criticals"),
])
def test_has_level(active_handler, logger, level, method):
    log = getattr(logger, level)
    log('Hello World')
    assert getattr(active_handler, method)


@pytest.mark.parametrize("level, method", [
    ("trace", "has_trace"),
    ("debug", "has_debug"),
    ("info", "has_info"),
    ("notice", "has_notice"),
    ("warning", "has_warning"),
    ("error", "has_error"),
    ("critical", "has_critical"),
])
def test_regex_matching(active_handler, logger, level, method):
    log = getattr(logger, level)
    log('Hello World')
    has_level_method = getattr(active_handler, method)
    assert has_level_method(re.compile('^Hello'))
    assert (not has_level_method(re.compile('world$')))
    assert (not has_level_method('^Hello World'))


def test_test_handler_cache(active_handler, logger):
    logger.warn('First line')
    assert len(active_handler.formatted_records) == 1
    # store cache, to make sure it is identifiable
    cache = active_handler.formatted_records
    assert len(active_handler.formatted_records) == 1
    assert cache is active_handler.formatted_records
    logger.warn('Second line invalidates cache')
    assert len(active_handler.formatted_records) == 2
    assert (cache is not active_handler.formatted_records)



import re
import sys

import logbook

import pytest

from .utils import capturing_stderr_context

__file_without_pyc__ = __file__
if __file_without_pyc__.endswith('.pyc'):
    __file_without_pyc__ = __file_without_pyc__[:-1]


def test_handler_exception(activation_strategy, logger):
    class ErroringHandler(logbook.TestHandler):

        def emit(self, record):
            raise RuntimeError('something bad happened')

    with capturing_stderr_context() as stderr:
        with activation_strategy(ErroringHandler()):
            logger.warn('I warn you.')
    assert 'something bad happened' in stderr.getvalue()
    assert 'I warn you' not in stderr.getvalue()


def test_formatting_exception():
    def make_record():
        return logbook.LogRecord('Test Logger', logbook.WARNING,
                                 'Hello {foo:invalid}',
                                 kwargs={'foo': 42},
                                 frame=sys._getframe())
    record = make_record()
    with pytest.raises(TypeError) as caught:
        record.message

    errormsg = str(caught.value)
    assert re.search(
        'Could not format message with provided arguments: Invalid '
        '(?:format specifier)|(?:conversion specification)|(?:format spec)',
        errormsg, re.M | re.S)
    assert "msg='Hello {foo:invalid}'" in errormsg
    assert 'args=()' in errormsg
    assert "kwargs={'foo': 42}" in errormsg
    assert re.search(
        r'Happened in file .*%s, line \d+' % re.escape(__file_without_pyc__),
        errormsg, re.M | re.S)



import logbook

import pytest

from .utils import capturing_stderr_context


def test_error_flag(logger):
    with capturing_stderr_context() as captured:
        with logbook.Flags(errors='print'):
            with logbook.Flags(errors='silent'):
                logger.warn('Foo {42}', 'aha')
        assert captured.getvalue() == ''

        with logbook.Flags(errors='silent'):
            with logbook.Flags(errors='print'):
                logger.warn('Foo {42}', 'aha')
        assert captured.getvalue() != ''

        with pytest.raises(Exception) as caught:
            with logbook.Flags(errors='raise'):
                logger.warn('Foo {42}', 'aha')
        assert 'Could not format message with provided arguments' in str(
            caught.value)


def test_disable_introspection(logger):
    with logbook.Flags(introspection=False):
        with logbook.TestHandler() as h:
            logger.warn('Testing')
            assert h.records[0].frame is None
            assert h.records[0].calling_frame is None
            assert h.records[0].module is None



import logbook

from .utils import capturing_stderr_context


def test_fingerscrossed(activation_strategy, logger, default_handler):
    handler = logbook.FingersCrossedHandler(default_handler,
                                            logbook.WARNING)

    # if no warning occurs, the infos are not logged
    with activation_strategy(handler):
        with capturing_stderr_context() as captured:
            logger.info('some info')
        assert captured.getvalue() == ''
        assert (not handler.triggered)

    # but if it does, all log messages are output
    with activation_strategy(handler):
        with capturing_stderr_context() as captured:
            logger.info('some info')
            logger.warning('something happened')
            logger.info('something else happened')
        logs = captured.getvalue()
        assert 'some info' in logs
        assert 'something happened' in logs
        assert 'something else happened' in logs
        assert handler.triggered


def test_fingerscrossed_factory(activation_strategy, logger):
    handlers = []

    def handler_factory(record, fch):
        handler = logbook.TestHandler()
        handlers.append(handler)
        return handler

    def make_fch():
        return logbook.FingersCrossedHandler(handler_factory,
                                             logbook.WARNING)

    fch = make_fch()
    with activation_strategy(fch):
        logger.info('some info')
        assert len(handlers) == 0
        logger.warning('a warning')
        assert len(handlers) == 1
        logger.error('an error')
        assert len(handlers) == 1
        assert handlers[0].has_infos
        assert handlers[0].has_warnings
        assert handlers[0].has_errors
        assert (not handlers[0].has_notices)
        assert (not handlers[0].has_criticals)
        assert (not handlers[0].has_debugs)

    fch = make_fch()
    with activation_strategy(fch):
        logger.info('some info')
        logger.warning('a warning')
        assert len(handlers) == 2


def test_fingerscrossed_buffer_size(activation_strategy):
    logger = logbook.Logger('Test')
    test_handler = logbook.TestHandler()
    handler = logbook.FingersCrossedHandler(test_handler, buffer_size=3)

    with activation_strategy(handler):
        logger.info('Never gonna give you up')
        logger.warn('Aha!')
        logger.warn('Moar!')
        logger.error('Pure hate!')

    assert test_handler.formatted_records == ['[WARNING] Test: Aha!',
                                              '[WARNING] Test: Moar!',
                                              '[ERROR] Test: Pure hate!']



# -*- coding: utf-8 -*-
from .utils import require_py3, capturing_stderr_context

import logbook


@require_py3
def test_default_format_unicode(logger):
    with capturing_stderr_context() as stream:
        logger.warn('\u2603')
    assert 'WARNING: testlogger: \u2603' in stream.getvalue()


@require_py3
def test_default_format_encoded(logger):
    with capturing_stderr_context() as stream:
        # it's a string but it's in the right encoding so don't barf
        logger.warn('\u2603')
    assert 'WARNING: testlogger: \u2603' in stream.getvalue()


@require_py3
def test_default_format_bad_encoding(logger):
    with capturing_stderr_context() as stream:
        # it's a string, is wrong, but just dump it in the logger,
        # don't try to decode/encode it
        logger.warn('Русский'.encode('koi8-r'))
    expected = "WARNING: testlogger: b'\\xf2\\xd5\\xd3\\xd3\\xcb\\xc9\\xca'"
    assert expected in stream.getvalue()


@require_py3
def test_custom_unicode_format_unicode(logger):
    format_string = ('[{record.level_name}] '
                     '{record.channel}: {record.message}')
    with capturing_stderr_context() as stream:
        with logbook.StderrHandler(format_string=format_string):
            logger.warn("\u2603")
    assert '[WARNING] testlogger: \u2603' in stream.getvalue()


@require_py3
def test_custom_string_format_unicode(logger):
    format_string = ('[{record.level_name}] '
                     '{record.channel}: {record.message}')
    with capturing_stderr_context() as stream:
        with logbook.StderrHandler(format_string=format_string):
            logger.warn('\u2603')
    assert '[WARNING] testlogger: \u2603' in stream.getvalue()


@require_py3
def test_unicode_message_encoded_params(logger):
    with capturing_stderr_context() as stream:
        logger.warn("\u2603 {0}", "\u2603".encode('utf8'))
    assert "WARNING: testlogger: \u2603 b'\\xe2\\x98\\x83'" in stream.getvalue()


@require_py3
def test_encoded_message_unicode_params(logger):
    with capturing_stderr_context() as stream:
        logger.warn('\u2603 {0}'.encode('utf8'), '\u2603')
    assert 'WARNING: testlogger: \u2603 \u2603' in stream.getvalue()



import sys

import logbook
import pytest

logbook.StderrHandler().push_application()


@pytest.fixture
def logger():
    return logbook.Logger('testlogger')


@pytest.fixture
def active_handler(request, test_handler, activation_strategy):

    s = activation_strategy(test_handler)
    s.activate()

    @request.addfinalizer
    def deactivate():
        s.deactivate()

    return test_handler


@pytest.fixture
def test_handler():
    return logbook.TestHandler()


class ActivationStrategy(object):

    def __init__(self, handler):
        super(ActivationStrategy, self).__init__()
        self.handler = handler

    def activate(self):
        raise NotImplementedError()  # pragma: no cover

    def deactivate(self):
        raise NotImplementedError()  # pragma: no cover

    def __enter__(self):
        self.activate()
        return self.handler

    def __exit__(self, *_):
        self.deactivate()


class ContextEnteringStrategy(ActivationStrategy):

    def activate(self):
        self.handler.__enter__()

    def deactivate(self):
        self.handler.__exit__(None, None, None)


class PushingStrategy(ActivationStrategy):

    def activate(self):
        from logbook.concurrency import is_gevent_enabled
        if is_gevent_enabled():
            self.handler.push_greenlet()
        else:
            self.handler.push_thread()

    def deactivate(self):
        from logbook.concurrency import is_gevent_enabled
        if is_gevent_enabled():
            self.handler.pop_greenlet()
        else:
            self.handler.pop_thread()


@pytest.fixture(params=[ContextEnteringStrategy, PushingStrategy])
def activation_strategy(request):
    return request.param


@pytest.fixture
def logfile(tmpdir):
    return str(tmpdir.join('logfile.log'))


@pytest.fixture
def default_handler(request):
    returned = logbook.StderrHandler()
    returned.push_application()
    request.addfinalizer(returned.pop_application)
    return returned

try:
    import gevent
except ImportError:
    pass
else:
    @pytest.fixture(scope="module", autouse=True, params=[False, True])
    def gevent(request):
        module_name = getattr(request.module, '__name__', '')
        if (not any(s in module_name for s in ('queues', 'processors'))
                and request.param):
            from logbook.concurrency import enable_gevent, _disable_gevent
            enable_gevent()

            @request.addfinalizer
            def fin():
                _disable_gevent()


def pytest_ignore_collect(path, config):
    if 'test_asyncio.py' in path.basename and (sys.version_info.major < 3 or sys.version_info.minor < 5):
        return True

    return False



# -*- coding: utf-8 -*-
"""
    logbook._fallback
    ~~~~~~~~~~~~~~~~~

    Fallback implementations in case speedups is not around.

    :copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
    :license: BSD, see LICENSE for more details.
"""
from itertools import count
from logbook.helpers import get_iterator_next_method
from logbook.concurrency import (
    thread_get_ident, greenlet_get_ident, thread_local, greenlet_local,
    ThreadLock, GreenletRLock, is_gevent_enabled, ContextVar, context_get_ident,
    is_context_enabled)

_missing = object()
_MAX_CONTEXT_OBJECT_CACHE = 256


def group_reflected_property(name, default, fallback=_missing):
    """Returns a property for a given name that falls back to the
    value of the group if set.  If there is no such group, the
    provided default is used.
    """
    def _get(self):
        rv = getattr(self, '_' + name, _missing)
        if rv is not _missing and rv != fallback:
            return rv
        if self.group is None:
            return default
        return getattr(self.group, name)

    def _set(self, value):
        setattr(self, '_' + name, value)

    def _del(self):
        delattr(self, '_' + name)
    return property(_get, _set, _del)


class _StackBound(object):

    def __init__(self, obj, push, pop):
        self.__obj = obj
        self.__push = push
        self.__pop = pop

    def __enter__(self):
        self.__push()
        return self.__obj

    def __exit__(self, exc_type, exc_value, tb):
        self.__pop()


class StackedObject(object):
    """Baseclass for all objects that provide stack manipulation
    operations.
    """

    def push_greenlet(self):
        """Pushes the stacked object to the greenlet stack."""
        raise NotImplementedError()

    def pop_greenlet(self):
        """Pops the stacked object from the greenlet stack."""
        raise NotImplementedError()

    def push_context(self):
        """Pushes the stacked object to the context stack."""
        raise NotImplementedError()

    def pop_context(self):
        """Pops the stacked object from the context stack."""
        raise NotImplementedError()

    def push_thread(self):
        """Pushes the stacked object to the thread stack."""
        raise NotImplementedError()

    def pop_thread(self):
        """Pops the stacked object from the thread stack."""
        raise NotImplementedError()

    def push_application(self):
        """Pushes the stacked object to the application stack."""
        raise NotImplementedError()

    def pop_application(self):
        """Pops the stacked object from the application stack."""
        raise NotImplementedError()

    def __enter__(self):
        if is_gevent_enabled():
            self.push_greenlet()
        else:
            self.push_thread()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if is_gevent_enabled():
            self.pop_greenlet()
        else:
            self.pop_thread()

    def greenletbound(self, _cls=_StackBound):
        """Can be used in combination with the `with` statement to
        execute code while the object is bound to the greenlet.
        """
        return _cls(self, self.push_greenlet, self.pop_greenlet)

    def contextbound(self, _cls=_StackBound):
        """Can be used in combination with the `with` statement to
        execute code while the object is bound to the concurrent
        context.
        """
        return _cls(self, self.push_context, self.pop_context)

    def threadbound(self, _cls=_StackBound):
        """Can be used in combination with the `with` statement to
        execute code while the object is bound to the thread.
        """
        return _cls(self, self.push_thread, self.pop_thread)

    def applicationbound(self, _cls=_StackBound):
        """Can be used in combination with the `with` statement to
        execute code while the object is bound to the application.
        """
        return _cls(self, self.push_application, self.pop_application)


class ContextStackManager(object):
    """Helper class for context objects that manages a stack of
    objects.
    """

    def __init__(self):
        self._global = []
        self._thread_context_lock = ThreadLock()
        self._thread_context = thread_local()
        self._greenlet_context_lock = GreenletRLock()
        self._greenlet_context = greenlet_local()
        self._context_stack = ContextVar('stack')
        self._cache = {}
        self._stackop = get_iterator_next_method(count())

    def iter_context_objects(self):
        """Returns an iterator over all objects for the combined
        application and context cache.
        """
        use_gevent = is_gevent_enabled()
        use_context = is_context_enabled()

        if use_gevent:
            tid = greenlet_get_ident()
        elif use_context:
            tid = context_get_ident()
        else:
            tid = thread_get_ident()

        objects = self._cache.get(tid)
        if objects is None:
            if len(self._cache) > _MAX_CONTEXT_OBJECT_CACHE:
                self._cache.clear()
            objects = self._global[:]
            objects.extend(getattr(self._thread_context, 'stack', ()))

            if use_gevent:
                objects.extend(getattr(self._greenlet_context, 'stack', ()))

            if use_context:
                objects.extend(self._context_stack.get([]))

            objects.sort(reverse=True)
            objects = [x[1] for x in objects]
            self._cache[tid] = objects
        return iter(objects)

    def push_greenlet(self, obj):
        self._greenlet_context_lock.acquire()
        try:
            # remote chance to conflict with thread ids
            self._cache.pop(greenlet_get_ident(), None)
            item = (self._stackop(), obj)
            stack = getattr(self._greenlet_context, 'stack', None)
            if stack is None:
                self._greenlet_context.stack = [item]
            else:
                stack.append(item)
        finally:
            self._greenlet_context_lock.release()

    def pop_greenlet(self):
        self._greenlet_context_lock.acquire()
        try:
            # remote chance to conflict with thread ids
            self._cache.pop(greenlet_get_ident(), None)
            stack = getattr(self._greenlet_context, 'stack', None)
            assert stack, 'no objects on stack'
            return stack.pop()[1]
        finally:
            self._greenlet_context_lock.release()

    def push_context(self, obj):
        self._cache.pop(context_get_ident(), None)
        item = (self._stackop(), obj)
        stack = self._context_stack.get(None)
        if stack is None:
            stack = [item]
            self._context_stack.set(stack)
        else:
            stack.append(item)

    def pop_context(self):
        self._cache.pop(context_get_ident(), None)
        stack = self._context_stack.get(None)
        assert stack, 'no objects on stack'
        return stack.pop()[1]

    def push_thread(self, obj):
        self._thread_context_lock.acquire()
        try:
            self._cache.pop(thread_get_ident(), None)
            item = (self._stackop(), obj)
            stack = getattr(self._thread_context, 'stack', None)
            if stack is None:
                self._thread_context.stack = [item]
            else:
                stack.append(item)
        finally:
            self._thread_context_lock.release()

    def pop_thread(self):
        self._thread_context_lock.acquire()
        try:
            self._cache.pop(thread_get_ident(), None)
            stack = getattr(self._thread_context, 'stack', None)
            assert stack, 'no objects on stack'
            return stack.pop()[1]
        finally:
            self._thread_context_lock.release()

    def push_application(self, obj):
        self._global.append((self._stackop(), obj))
        self._cache.clear()

    def pop_application(self):
        assert self._global, 'no objects on application stack'
        popped = self._global.pop()[1]
        self._cache.clear()
        return popped



import pickle
import sys

import logbook
from logbook.helpers import iteritems, xrange, u

import pytest


def test_basic_logging(active_handler, logger):
    logger.warn('This is a warning.  Nice hah?')

    assert active_handler.has_warning('This is a warning.  Nice hah?')
    assert active_handler.formatted_records == [
        '[WARNING] testlogger: This is a warning.  Nice hah?']


def test_exception_catching(active_handler, logger):
    assert not active_handler.has_error()
    try:
        1 / 0
    except Exception:
        logger.exception()
    try:
        1 / 0
    except Exception:
        logger.exception('Awesome')
    assert active_handler.has_error('Uncaught exception occurred')
    assert active_handler.has_error('Awesome')
    assert active_handler.records[0].exc_info is not None
    assert '1 / 0' in active_handler.records[0].formatted_exception


def test_exception_catching_with_unicode():
    """ See https://github.com/getlogbook/logbook/issues/104
    """
    try:
        raise Exception(u('\u202a test \u202c'))
    except:
        r = logbook.LogRecord('channel', 'DEBUG', 'test',
                              exc_info=sys.exc_info())
    r.exception_message


@pytest.mark.parametrize('as_tuple', [True, False])
def test_exc_info(as_tuple, logger, active_handler):
    try:
        1 / 0
    except Exception:
        exc_info = sys.exc_info()
        logger.info("Exception caught",
                    exc_info=exc_info if as_tuple else True)
    assert active_handler.records[0].exc_info is not None
    assert active_handler.records[0].exc_info == exc_info


def test_to_dict(logger, active_handler):
    try:
        1 / 0
    except Exception:
        logger.exception()
        record = active_handler.records[0]

    exported = record.to_dict()
    record.close()
    imported = logbook.LogRecord.from_dict(exported)
    for key, value in iteritems(record.__dict__):
        if key[0] == '_':
            continue
        assert value == getattr(imported, key)


def test_pickle(active_handler, logger):
    try:
        1 / 0
    except Exception:
        logger.exception()
        record = active_handler.records[0]
    record.pull_information()
    record.close()

    for p in xrange(pickle.HIGHEST_PROTOCOL):
        exported = pickle.dumps(record, p)
        imported = pickle.loads(exported)
        for key, value in iteritems(record.__dict__):
            if key[0] == '_':
                continue
            imported_value = getattr(imported, key)
            if isinstance(value, ZeroDivisionError):
                # in Python 3.2, ZeroDivisionError(x) != ZeroDivisionError(x)
                assert type(value) is type(imported_value)
                assert value.args == imported_value.args
            else:
                assert value == imported_value



import sys
import logbook


class MyObject(object):
    def __init__(self, logger_func):
        self._logger_func = logger_func

    def __str__(self):
        self._logger_func("this debug message produced in __str__")
        return "<complex object>"


class FakeLock(object):
    def __init__(self):
        self._acquired = False
        self._deadlock_occurred = False

    def acquire(self):
        if self._acquired:
            self._deadlock_occurred = True
        self._acquired = True

    def release(self):
        self._acquired = False


def test_deadlock_in_emit():
    logbook_logger = logbook.Logger("logbook")
    obj = MyObject(logbook_logger.info)
    stream_handler = logbook.StreamHandler(stream=sys.stderr,
                                           level=logbook.DEBUG)
    stream_handler.lock = FakeLock()
    with stream_handler.applicationbound():
        logbook_logger.info("format this: {}", obj)
    assert not stream_handler.lock._deadlock_occurred



import pytest
import logbook
import asyncio
from logbook.concurrency import has_contextvars

ITERATIONS = 100


@pytest.mark.skipif(not has_contextvars, reason="Contexvars not available")
def test_asyncio_context_management(logger):
    h1 = logbook.TestHandler()
    h2 = logbook.TestHandler()

    async def task(handler, msg):
        for _ in range(ITERATIONS):
            with handler.contextbound():
                logger.info(msg)

            await asyncio.sleep(0)  # allow for context switch

    asyncio.get_event_loop().run_until_complete(asyncio.gather(task(h1, 'task1'), task(h2, 'task2')))

    assert len(h1.records) == ITERATIONS
    assert all(['task1' == r.msg for r in h1.records])

    assert len(h2.records) == ITERATIONS
    assert all(['task2' == r.msg for r in h2.records])



from textwrap import dedent

import logbook

from .utils import make_fake_mail_handler


def test_handler_filter_after_processor(activation_strategy, logger):
    handler = make_fake_mail_handler(
        format_string=dedent('''
            Subject: Application Error for {record.extra[path]} [{record.extra[method]}]

            Message type:       {record.level_name}
            Location:           {record.filename}:{record.lineno}
            Module:             {record.module}
            Function:           {record.func_name}
            Time:               {record.time:%Y-%m-%d %H:%M:%S}
            Remote IP:          {record.extra[ip]}
            Request:            {record.extra[path]} [{record.extra[method]}]

            Message:

            {record.message}
            ''').lstrip(),
        filter=lambda r, h: 'ip' in r.extra,
        bubble=False)

    class Request(object):
        remote_addr = '127.0.0.1'
        method = 'GET'
        path = '/index.html'

    def handle_request(request):
        def inject_extra(record):
            record.extra['ip'] = request.remote_addr
            record.extra['method'] = request.method
            record.extra['path'] = request.path

        processor = logbook.Processor(inject_extra)
        with activation_strategy(processor):
            handler.push_thread()
            try:
                try:
                    1 / 0
                except Exception:
                    logger.exception('Exception happened during request')
            finally:
                handler.pop_thread()

    handle_request(Request())
    assert len(handler.mails) == 1
    mail = handler.mails[0][2]
    assert 'Subject: Application Error for /index.html [GET]' in mail
    assert '1 / 0' in mail


def test_handler_processors(activation_strategy, logger):
    handler = make_fake_mail_handler(
        format_string=dedent('''
            Subject: Application Error for {record.extra[path]} [{record.extra[method]}]

            Message type:       {record.level_name}
            Location:           {record.filename}:{record.lineno}
            Module:             {record.module}
            Function:           {record.func_name}
            Time:               {record.time:%Y-%m-%d %H:%M:%S}
            Remote IP:          {record.extra[ip]}
            Request:            {record.extra[path]} [{record.extra[method]}]

            Message:

            {record.message}
            ''').lstrip())

    class Request(object):
        remote_addr = '127.0.0.1'
        method = 'GET'
        path = '/index.html'

    def handle_request(request):
        def inject_extra(record):
            record.extra['ip'] = request.remote_addr
            record.extra['method'] = request.method
            record.extra['path'] = request.path

        processor = logbook.Processor(inject_extra)
        with activation_strategy(processor):
            handler.push_thread()
            try:
                try:
                    1 / 0
                except Exception:
                    logger.exception('Exception happened during request')
            finally:
                handler.pop_thread()

    handle_request(Request())
    assert len(handler.mails) == 1
    mail = handler.mails[0][2]
    assert 'Subject: Application Error for /index.html [GET]' in mail
    assert '1 / 0' in mail
